#!/usr/bin/env python2.4
#
# tidybot.py

"""
Utility program for checking all XHTML files in a directory tree
for errors, and formatting the results as a webpage.

Main program module.
"""

__author__ = "Leo Breebaart <leo@kronto.org>"
__version__ = "1.6.1"
__date__ = "5 November 2005"


import sys
import os.path
import imp
import time
import fnmatch
import inspect
import shutil
import optparse
import urllib
import subprocess


# We don't actually use ctypes ourselves, but 'tidy' depends on
# it, so we try to import it here in order to be able to generate
# a more precise error message.
try:
    import ctypes
except ImportError:
    print "Error: Python module 'ctypes' not found"
    print "Get it from <http://starship.python.net/crew/theller/ctypes/>"
    sys.exit(1)

try:
    import tidy
except ImportError:
    print "Error: Python module 'tidy' not found"
    print "Get it from <http://utidylib.berlios.de/>"
    sys.exit(1)

try:
    from HTMLTemplate import Template
except ImportError:
    print "Error: Python module not found: HTMLTemplate"
    print "Get it from <http://freespace.virgin.net/hamish.sanderson/htmltemplate.html>"
    sys.exit(1)


# These two functions exist solely so that py2exe will be able to
# find the tidybot data files correctly. (Relative pathnames do
# not work as expected if your execution context is a zipfile, as
# it is when running from a py2exe-generated executable.
def main_is_frozen():
    return (hasattr(sys, "frozen") or # new py2exe
            hasattr(sys, "importers") # old py2exe
            or imp.is_frozen("__main__")) # tools/freeze

def get_main_dir():
    if main_is_frozen():
        return os.path.dirname(sys.executable)
    return os.path.dirname(inspect.getabsfile(get_main_dir))


# Function taken from the Python Cookbook, recipe 4.18
#
def listfiles(root, patterns='*', recurse=1, return_folders=0):
    """ A user-friendly wrapper around os.path.walk(). """

    # Expand patterns from semicolon-separated string to list
    pattern_list = patterns.split(',')
    class Bunch:
        """Collect input and output arguments into one bunch"""
        def __init__(self, **kwds): self.__dict__.update(kwds)
    arg = Bunch(recurse=recurse, pattern_list=pattern_list,
        return_folders=return_folders, results=[])

    def visit(arg, dirname, files):
        # Append to arg.results all relevant files (and perhaps folders)
        for name in files:
            fullname = os.path.normpath(os.path.join(dirname, name))
            if arg.return_folders or os.path.isfile(fullname):
                for pattern in arg.pattern_list:
                    if fnmatch.fnmatch(name, pattern):
                        arg.results.append(fullname)
                        break
        # Block recursion if recursion was disallowed
        if not arg.recurse: files[:] = []

    os.path.walk(root, visit, arg)

    return arg.results


def read_lines(filename):
    """ 
    Read and return a list of strings (one per line) from
    'filename'. Empty lines and comments (lines starting with
    '#') are discarded.
    """
    infile = open(filename)
    lines = infile.read().splitlines()
    infile.close()

    result = []
    for line in lines:
        if not line or line.startswith('#'):
            continue
        result.append(line)
    return result


class TidybotError(Exception):
    """
    Raised if a problem during a Tidybot run is
    encountered.
    """
    label = "Tidybot Error"
    def __init__(self, msg):
        Exception.__init__(self)
        self.msg = msg

    def __str__(self):
        return self.msg


class TidybotOptionError(TidybotError):
    """
    Raised if a problem with a Tidybot input option is
    encountered.
    """
    label = "Tidybot Option Error"


class TidyOptionError(TidybotError):
    """
    Raised if an unparseable Tidy config file line is
    encountered.
    """
    label = "Tidy Option Error"


def parse_tidy_options(sequence):
    """ Read and return a dictionary of tidy options (one per
    line) from 'sequence'. The format is 'key:value' as in
    conventional tidy config files."""
    tidy_options = {}
    for pair in sequence:
        try:
            key,value = pair.lower().split(':')
        except ValueError:
            raise TidyOptionError("Not a valid Tidy 'key:value' option: '%s'" % pair)
        tidy_options[key.strip()] = value.strip()
    return tidy_options


def found_in(stringlist, target):
    """ Returns true if at least one of the strings in
    'stringlist' is contained in the string 'target'"""
    for src in stringlist:
        if src in target:
            return True
    return False


class FileBlock:
    """ Convenience class for collecting the results of running
    tidy on a single file."""

    def __init__(self, longfilename, results, config, sourceroot):
        self.config = config
        self.longfilename = longfilename
	self.filename_as_url = urllib.pathname2url(longfilename)
        self.filename = self.longfilename.replace(sourceroot, "")
        self.results = list(results)
        self.nerrors = self.nwarnings = 0
        self.nsuppressions = 0
        for res in results:
            if found_in(config.suppressions, str(res)):
                self.nsuppressions += 1
                self.results.remove(res)
            else:
                if "Error:" in str(res):
                    self.nerrors += 1
                elif "Warning:" in str(res):
                    self.nwarnings += 1


class SessionInfo:
    """ Convenience class for storing the cumulative results of
    running tidy on an entire directory tree."""

    def __init__(self, rootdir, fileblocks, config):
        self.config = config
        self.rootdir = rootdir
        self.nfiles = len(fileblocks)
        self.nerrors = sum(fb.nerrors for fb in fileblocks)
        self.nwarnings = sum(fb.nwarnings for fb in fileblocks)
        self.nsuppressions = sum(fb.nsuppressions for fb in fileblocks)
        self.fileblocks = [fb for fb in fileblocks if fb.results]
        self.nproblems = len(self.fileblocks)
        if self.nfiles:
            self.ratio = 100.0 * self.nproblems / self.nfiles
        else:
            self.ratio = 0
        self.duration = None

        basename = os.path.basename(config.output_filename)
        dirname = os.path.dirname(config.output_filename)
        if basename.endswith(".html"):
            basename = basename[:-5]
        rootstring = rootdir.replace(os.path.sep, '.')
        rootstring = rootstring.replace(':', '')
        if rootstring[0] == '.':
            rootstring = rootstring[1:]
        self.listingname = "%s-%shtml" % (basename, rootstring)
        self.listingname = os.path.join(dirname, self.listingname)


def format_duration(duration):
    """ Format 'duration' seconds into a more human-readable
    hh:mm:ss format.
    """
    hours = duration // 3600
    minutes = (duration % 3600) // 60
    seconds = (duration % 3600) % 60
    if hours:
        duration_str = ("%d:%2d:%02d"
                        % (hours, minutes, seconds))
    else:
        duration_str = ("%2d:%02d"
                        % (minutes, seconds))
    return duration_str


def render_errors(templatename, config, sourceroot, fileblocks):
    """ Use result list 'fileblocks' to fill file 'templatename',
    the Tidybot subpage listing the actual problems encountered."""

    templatefile = open(templatename)
    html = templatefile.read()
    templatefile.close()

    def render_template(node, fileblocks):
        node.version.content = __version__
        node.date.content = time.asctime()
        node.mainpage.atts['href'] = os.path.basename(config.output_filename)
        node.sourceroot.content = sourceroot
        node.stylesheet.atts['href'] = config.stylesheet
        node.fileblock.repeat(render_fileblock, fileblocks)

    def render_fileblock(node, fileblock):
        node.filename.content = fileblock.filename
        node.filename.atts['href'] = fileblock.filename_as_url
        node.listitem.repeat(render_listitem, fileblock.results)

    def render_listitem(listitem, result):
        listitem.content = "%s" % result
        if "Error:" in str(result):
            listitem.atts['class'] = "error"
        else:
            listitem.atts['class'] = "warning"

    template = Template(render_template, html)
    return template.render(fileblocks)


def render_main(templatename, config, sessions_info):
    """ Use list of class instance 'sessioninfo' to fill file
    'templatename', the file containing the Tidybot main page."""

    templatefile = open(templatename)
    html = templatefile.read()
    templatefile.close()

    def render_template(node, sessions_info):
        node.version.content = __version__
        node.date.content = time.asctime()
        node.stylesheet.atts['href'] = config.stylesheet

        total_nfiles = sum(si.nfiles for si in sessions_info)
        total_nerrors = sum(si.nerrors for si in sessions_info)
        total_nwarnings = sum(si.nwarnings for si in sessions_info)
        total_nsuppressions = sum(si.nsuppressions for si in sessions_info)
        total_duration = sum(si.duration for si in sessions_info)
        allroots = [si.rootdir for si in sessions_info]

        node.nfiles.content = "%d" % total_nfiles
        node.nerrors.content = "%d" % total_nerrors
        node.nwarnings.content = "%d" % total_nwarnings
        node.nsuppressions.content = "%d" % total_nsuppressions
        node.duration.content = "%s" % format_duration(total_duration)
        node.sessions.repeat(render_session, sessions_info)
        node.sourceroots.repeat(render_sourceroot, allroots)
        node.exclusions.repeat(render_exclusion, config.exclusions)
        node.suppressions.repeat(render_suppression, config.suppressions)
        node.tidy_options.repeat(render_tidy_option, config.tidy_options.items())

    def render_session(node, sessioninfo):
        node.rootdir.content = sessioninfo.rootdir
        node.rootdir.atts['href'] = "%s" % os.path.basename(sessioninfo.listingname)
        node.nproblems.content = "%s" % sessioninfo.nproblems
        node.nfiles2.content = "%d" % sessioninfo.nfiles
        node.ratio.content = "%.1f%%" % sessioninfo.ratio

    def render_sourceroot(sourceroot, value):
        sourceroot.content = '"%s"' % value

    def render_exclusion(exclusion, value):
        exclusion.content = '"%s"' % value

    def render_suppression(suppression, value):
        suppression.content = '"%s"' % value

    def render_tidy_option(tidy_option, value):
        tidy_option.content = '"%s = %s"' % value

    template = Template(render_template, html)
    return template.render(sessions_info)


def merge_errors(errlist1, errlist2):
    result = []
    for i in errlist1+errlist2:
        if not(str(i)) in result:
            result.append(str(i))

    return result

def collect_errors(sourceroot, config, update_function = None):
    """ Recursively descend through the directory 'sourceroot' and
    collect error information about all the HTML files
    found."""

    start = time.time()

    filelist = listfiles(sourceroot, ','.join(config.inclusions))

    # Disregard the files matched by the optional global
    # exclusions list.
    for excl in config.exclusions:
        filelist = [f for f in filelist if excl not in f]

    # Store results for each file in a FileBlock class instance
    fileblocks = []
    for filename in filelist:
        if config.verbose:
            print "Checking %s:" % filename
        errors = tidy.parse(filename, **config.tidy_options).errors

        # Check if a second pass as an XML checker has been specified:
        if config.xml_pass:
            xml_options = dict(config.tidy_options)
            xml_options["input-xml"] = "true"
            xml_errors = tidy.parse(filename, **xml_options).errors
            errors = merge_errors(errors, xml_errors)

        fileblocks.append(FileBlock(filename, errors, config, sourceroot))
        if config.verbose:
            for err in errors:
                print "   %s" % err
        if update_function:
            config.current_count += 1
            update_function(config.current_count, config.max_count)

    # Collect info for the entire session
    sessioninfo = SessionInfo(sourceroot, fileblocks, config)

    stop = time.time()
    delta = stop - start
    sessioninfo.duration = delta

    return sessioninfo



def send_mail_report(mailto, note, sessions_info):

    sendmail = os.getenv("SENDMAIL", "sendmail")

    total_nfiles = sum(si.nfiles for si in sessions_info)
    total_nerrors = sum(si.nerrors for si in sessions_info)
    total_nwarnings = sum(si.nwarnings for si in sessions_info)
    allroots = "\n  ".join(si.rootdir for si in sessions_info)

    msg = """\
To: %s
Subject: Tidybot results

Tidybot results for:
  %s

Files checked:     %d
Errors detected:   %d
Warnings detected: %d
""" % (mailto, allroots, total_nfiles, total_nerrors, total_nwarnings)

    if note:
        msg = """%s\

User-supplied note:
  %s
""" % (msg, note)

    try:
        pipe = subprocess.Popen([sendmail, "-oi", "-t"],
                                stdin=subprocess.PIPE,
                                stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE)
    except OSError, eeks:
	raise TidybotError('Sendmail call failed with operating system error:\n"%s"\n(Note: --mailto option will not work on Windows)' % str(eeks)) 
    output, errors = pipe.communicate(msg)
    result = pipe.returncode
    if result < 0:
	raise TidybotError("Sendmail call failed. Result = %d\n(Note: --mailto option will not work on Windows)" % result)


def run_tidybot(config, update_function = None):
    """ Run tidybot. Configuration parameters are found in the
    global Config class object."""

    def write_page(pagename, rendered_html):
        """ Write renderered_html to an output file."""
        outfile = open(pagename, 'w')
        outfile.write(rendered_html + "\n")
        outfile.close()

    if config.verbose:
        print "*** Starting Tidybot %s in verbose mode" % __version__
        print "Main output file: %s" % config.output_filename
        print "Source directories:"
        for i in config.sourceroots:
            print "   %s" % i
        if config.inclusions:
            print "Inclusions from %s:" % config.include_from
            for i in config.inclusions:
                print "   %s" % i
        if config.exclusions:
            print "Exclusions from %s:" % config.exclude_from
            for i in config.exclusions:
                print "   %s" % i
        if config.suppressions:
            print "Suppressions from %s:" % config.suppress_from
            for i in config.suppressions:
                print "   %s" % i
        if config.tidy_options:
            print "Tidy options from %s:" % config.tidy_options_from
            for i in config.tidy_options.items():
                print "   %s:%s" % i

    sessions_info = {}
    for root in config.sourceroots:
        session_info = collect_errors(root, config, update_function)
        sessions_info[root] = session_info

    main_pagename = config.output_filename

    package_dir = get_main_dir()
    main_template = os.path.join(package_dir, 'templates', 'tidybot.tmpl')
    main_html = render_main(main_template, config, sessions_info.values())
    write_page(main_pagename, main_html)

    for root in config.sourceroots:
        info = sessions_info[root]
        errors_pagename = info.listingname
        errors_template = os.path.join(package_dir, 'templates', 'tidybot-listing.tmpl')
        errors_html = render_errors(errors_template, config,
                                    root, info.fileblocks)
        write_page(errors_pagename, errors_html)

    if config.stylesheet == config.DEFAULT_STYLESHEET:
        css_file = os.path.join(package_dir, 'css', 'tidybot.css')
        png_file = os.path.join(package_dir, 'css', 'tidybot-bg.png')
        target_dir = os.path.dirname(main_pagename)
        shutil.copy(css_file, target_dir)
        shutil.copy(png_file, target_dir)

    if config.mailto:
        send_mail_report(config.mailto, config.note, sessions_info.values())


class TidybotOptionParser(optparse.OptionParser):

    def error(self, msg):
        if msg:
            sys.stderr.write("error: %s\n\n" % msg)
        self.print_help(sys.stderr)
        self.exit(2)


class Config(object):
    """ Hold configuration info supplied on or derived from
    command-line options.
    """
    DEFAULT_OUTPUT_DIRECTORY = ''
    DEFAULT_INCLUSIONS = "*.html,*.htm"
    DEFAULT_STYLESHEET = "tidybot.css"

    def __init__(self, options, args):
        self.verbose = False
        self.xml_pass = False
        self.sourceroots = []
        self.inclusions = set()
        self.exclusions = set()
        self.suppressions = set()
        self.tidy_options = {}
	self.include_from = None
        self.exclude_from = None
        self.suppress_from = None
        self.tidy_options_from = None
	self.include = self.DEFAULT_INCLUSIONS
        self.exclude = None
        self.suppress = None
        self.tidy_options_str = None
        self.stylesheet = self.DEFAULT_STYLESHEET
        self.output_directory = self.DEFAULT_OUTPUT_DIRECTORY
        self.output_filename = "tidybot.html"
        self.mailto = None
        self.note = None
        self.debug = False

        self.__dict__.update(options)

        if self.note and not self.mailto:
            raise TidybotOptionError("-n/--note option requires --mailto option")

        self.output_directory = os.path.abspath(self.output_directory)
        self.output_filename = os.path.join(self.output_directory, "tidybot.html")
        check_output_file(self.output_filename)

        self.exclusions.update(exclusions_from_file(self.exclude_from))
        self.suppressions.update(suppressions_from_file(self.suppress_from))
        self.tidy_options.update(tidy_options_from_file(self.tidy_options_from))
        self.inclusions.update(inclusions_from_file(self.include_from))

        self.exclusions.update(exclusions_from_string(self.exclude))
        self.suppressions.update(suppressions_from_string(self.suppress))
        self.tidy_options.update(tidy_options_from_string(self.tidy_options_str))
        self.inclusions.update(inclusions_from_string(self.include))

        self.max_count = 0

        if not args:
            raise TidybotOptionError("Missing 'source directory' argument")

        for sourceroot in args:
            if not os.path.exists(sourceroot):
                raise TidybotOptionError("Source directory does not exist: %s" % sourceroot)
            if not os.path.isdir(sourceroot):
                raise TidybotOptionError("Source is not a directory: %s" % sourceroot)
            if not sourceroot.endswith(os.path.sep):
                sourceroot += os.path.sep
	    self.sourceroots.append(sourceroot)

        for sourceroot in self.sourceroots:
            filelist = listfiles(sourceroot, ','.join(self.inclusions))
            for excl in self.exclusions:
                filelist = [f for f in filelist if excl not in f]
            self.max_count += len(filelist)
        self.current_count = 0

    def __str__(self):
        result = ""
        for key,value in self.__dict__.iteritems():
            result += "%s: %s\n" % (key,value)
        return result


def check_output_file(filename):
    """ Test if file 'filename' can be written to."""
    try:
        testfile = open(filename, "w")
        testfile.close()
    except IOError, eeks:
        raise TidybotOptionError("Could not open output file '%s':\n  %s" % (filename, eeks))


def exclusions_from_file(filename):
    if not filename:
        return set()
    try:
        exclusions = set(read_lines(filename))
    except IOError, eeks:
        raise TidybotOptionError("Could not open exclusions file '%s':\n  %s" % (filename, eeks))
    return exclusions


def inclusions_from_file(filename):
    if not filename:
        return set()
    try:
        inclusions = set(read_lines(filename))
    except IOError, eeks:
        raise TidybotOptionError("Could not open inclusions file '%s':\n  %s" % (filename, eeks))
    return inclusions


def suppressions_from_file(filename):
    if not filename:
        return set()
    try:
        suppressions = set(read_lines(filename))
    except IOError, eeks:
        raise TidybotOptionError("Could not open suppressions file '%s':\n  %s" % (filename, eeks))
    return suppressions


def tidy_options_from_file(filename):
    if not filename:
        return dict()
    try:
        tidy_options = parse_tidy_options(read_lines(filename))
    except IOError, eeks:
        raise TidybotOptionError("Could not open tidy config file '%s':\n  %s" % (filename, eeks))
    return tidy_options


def inclusions_from_string(incl_str):
    if not incl_str:
        return set()
    inclusions = incl_str.split(',')
    inclusions = set(i.strip() for i in inclusions)
    inclusions = set(i for i in inclusions if i)
    return inclusions


def exclusions_from_string(excl_str):
    if not excl_str:
        return set()
    exclusions = excl_str.split(',')
    exclusions = set(e.strip() for e in exclusions)
    exclusions = set(e for e in exclusions if e)
    return exclusions


def suppressions_from_string(suppr_str):
    if not suppr_str:
        return set()
    suppressions = suppr_str.split(',')
    suppressions = set(s.strip() for s in suppressions)
    suppressions = set(s for s in suppressions if s)
    return suppressions


def tidy_options_from_string(opt_str):
    if not opt_str:
        return dict()
    option_pairs = opt_str.lower().split(',')
    tidy_options = parse_tidy_options(option_pairs)
    return tidy_options


def version_string():
    return "Tidybot %s, %s\n(using TidyLib release %s)\nWritten by %s" \
           % (__version__, __date__,
              ctypes.string_at(tidy.lib._tidy.ReleaseDate()),
              __author__)


def process_commandline(argv):
    """ Check and process all the commandline options supplied by
    the user.
    """

    parser = TidybotOptionParser(prog = "tidybot",
                                 usage = "tidybot [OPTION]... <source directory>...",
                                 version = version_string(),
                                 formatter=optparse.TitledHelpFormatter())

    parser.add_option("-v", "--verbose", action="store_true",
                      help="turn on verbose mode")
    parser.add_option("-x", "--xml-pass", action="store_true",
                      help="perform a second, xml-checking pass")
    parser.add_option("--output-directory",
                      metavar="DIR", default=Config.DEFAULT_OUTPUT_DIRECTORY,
                      help="place results in DIR")

    parser.add_option("-i", "--include",
                      metavar="LIST", default=Config.DEFAULT_INCLUSIONS,
                      help="include filenames matching patterns from comma-separated LIST")
    parser.add_option("--include-from",
                      metavar="FILE",
                      help="include filenames matching patterns from FILE")

    parser.add_option("-e", "--exclude",
                      metavar="LIST",
                      help="exclude filenames containing strings from comma-separated LIST")
    parser.add_option("--exclude-from",
                      metavar="FILE",
                      help="exclude filenames containing strings from FILE")

    parser.add_option("-s", "--suppress",
                      metavar="LIST",
                      help="suppress results containing strings from comma-separated LIST")
    parser.add_option("--suppress-from",
                      metavar="FILE",
                      help="suppress results containing strings from FILE")

    parser.add_option("-t", "--tidy-options",
                      dest="tidy_options_str", metavar="LIST",
                      help="use tidy options listed in comma-separated LIST")
    parser.add_option("--tidy-options-from",
                      metavar="FILE",
                      help="use tidy config file FILE")
    parser.add_option("--stylesheet", metavar="URL", default=Config.DEFAULT_STYLESHEET,
                      help="reference the stylesheet at this URL", )
    parser.add_option("--debug", action="store_true", help=optparse.SUPPRESS_HELP)
    parser.add_option("--mailto", metavar="ADDRESS",
                      help="mail brief synopsis to ADDRESS when done", )
    parser.add_option("-n", "--note", metavar="NOTE",
                      help="include NOTE in the --mailto message", )

    opts,args = parser.parse_args(argv[1:])

    try:
        config = Config(vars(opts), args)
    except TidybotOptionError, eeks:
        raise parser.error(eeks)
    except TidyOptionError, eeks:
        raise parser.error(eeks)

    if config.debug:
        print config

    return config


def main(argv=None):
    if argv is None:
        argv = sys.argv

    config = process_commandline(argv)

    try:
        run_tidybot(config)
    except tidy.error.TidyLibError, eeks:
        print "The Tidy library reported an error:"
        print "  %s" % eeks
        return 1

    return 0


if __name__ == "__main__":

    sys.exit(main())
