# wingwebbrowser.py - an improved version of Python's own webbrowser.py module.
#
# This patched module was submitted by Stephen Deibel to the Python developer's
# list for the Python 2.4 track. It is not a perfect version but we are using
# it in VISAN because the Python 2.3 version *really* does not work very well.
#
# We have applied some changes of our own for improved MacOSX compatibility.
#
#
# Relevant meta-info:
#
# URL: <http://mail.python.org/pipermail/patches/2003-April/012263.html>
#
# Category: Library (Lib)
# Group: Python 2.3
# Status: Open
# Resolution: None
# Priority: 5
# Submitted By: Stephan R.A. Deibel (sdeibel)
# Assigned to: Nobody/Anonymous (nobody)
# Summary: Multiple webbrowser.py bug fixes / improvements
#
# Initial Comment: In using webbrowser.py we uncovered and fixed
# a number of problems and made some improvements in usability
# and consistency of behavior.
#
# Appended below is a summary of the changes made. This list is
# also found at the top of the uploaded file, which was based on
# the version of webbrowser.py found in Python 2.3a2. Sorry to
# submit so many changes at once but they were all made in a
# period of a few days when we went through this module for use
# in Wing IDE.
#
# I'm also submitting some unit tests for the module.
#
# I've tried to review everything carefully, further review is
# definately needed. Feel free to contact me if you have
# questions or comments or want me to make changes and resubmit.
#
# Hope this is helpful.
#
# - Stephan
#
#
# Stephan Deibel
#
# --
# Wing IDE for Python
# Archaeopteryx Software, Inc
# Take Flight!
#
# www.wingide.com
#
"""Interfaces for launching and remotely controlling Web browsers.

This is based on Python2.3a2's webbrowser.py module with some changes
made at Archaeopteryx Software, Inc.  For more info write to us at
support@wingide.com.

Bugs fixed:

* Don't apply lower() to command lines or commands that are going to be executed
* Don't confuse browser name/id with command line used to launch the browser
* Require that browser commands be executable
* Identify user-provided strings as a command line by looking for any args and
  not just for %s
* Handle spaces in user-provided command names, either if quoted on the
  user-provided command line or escaped with back slashes
* Use '%s' instead of "%s", which avoids character interpretation on
  Unix command lines
* Escape and quote urls more safely to prevent crafting urls that result
  in command lines that execute arbitrary commands
* Fixed Galeon so it doesn't hang up the app until the browser is quit
* Added the Mac OS X support that doesn't make bad environment assumptions
  on OS 10.1
* Fixed win32 use of Netscape, which before would never work because
  the _tryorder entry was being pruned out
* Leave found items in _browsers and _tryorder on OS X (but prepend
  the OS X specific support in _tryorder).  OS X is Unix so these are useful
  there.
* Now add %s to end of command lines if they don't contain %s already, for
  compatibility e.g. with KDE's default value for BROWSER environment
* Now add '&' to end of Unix command lines that are going to be launched
  from GenericBrowser to avoid hanging up on user-provided command lines
* Added additional quoting of command names, so some of the browser classes
  can work with a renamed browser with a space in the executable name
* Added a number of return values / checks where before there were bad
  assumptions in the code that might have led to errors

Other internal changes made:

* Added unit tests
* Removed the loop in get() which always returned in the first iteration
* _tryorder more tightly coupled with registering in _browsers, and removed
  potentially problematic last-ditch GenericBrowser registering clause and
  the prune-_tryorder clauses from the end of the module
* _iscommand also returns true if cmd is an existing file name, and it
  returns the full path or None instead of just True and False
* _synthesize registers and returns GenericBrowser if synthesis fails but
  the user-provided command line looks valid
* Some additional uses of True and False instead of 1 and 0

Changes to the public interface:

* Added optional update_tryorder arg to register() (default value is same
  as previous action of this function)
* Open/open_new return False if command definately failed and True if
  it may have succeeded (both previously returned None in all cases)
* Get() returns a GenericCommand if 'using' is given and doesn't name or
  synthesize a browser (previously it did this correctly only when %s
  was in the given command line)
* Values passed via BROWSER environment will be registered and synthesized
  in the same way as 'using' arg values are treated in get()

"""

import os
import sys
import stat

__all__ = ["Error", "open", "get", "register"]

try:
    if True: pass
except:
    True = 1
    False = 0

class Error(Exception):
    pass

_browsers = {}          # Dictionary of available browser controllers
_tryorder = []          # Preference order of available browsers

def register(name, klass, instance=None, update_tryorder=1):
    """Register a browser connector and, optionally, connection."""
    _browsers[name.lower()] = [klass, instance]
    if update_tryorder > 0:
        _tryorder.append(name)
    elif update_tryorder < 0:
        _tryorder.insert(0, name)

def get(using=None):
    """Return browser instance for first registered browser or for the
    given browser when 'using' is not None.  If using is specified but
    the browser for it cannot be found, then using is assumed to be a
    command line and a temporary (unregistered) generic browser for it
    is returned."""

    if using is not None:
        browser = using
    else:
        if len(_tryorder) == 0:
            raise Error("could not locate runnable browser")
        browser = _tryorder[0]

    try:
        command = _browsers[browser.lower()]
    except KeyError:
        command = _synthesize(browser)
    if command[1] is None:
        if command[0] is None:
            raise Error("could not locate runnable browser")
        return command[0]()
    else:
        return command[1]

# Please note: the following definition hides a builtin function.
def open(url, new=False, autoraise=True):
    """Open given url with first registered browser.  Returns False if
    the command definately failed and True it if may have succeeded."""
    return get().open(url, new, autoraise)

def open_new(url):
    """Open given url with first registered browser, creating a new window.
    Returns False if the command definately failed and True if it may have
    succeeded."""
    return get().open(url, True)


def _cmdline_browser(cmdline):
    """Get a generic browser for given command line"""
    cmdname, args = _splitcommand(cmdline)
    if _iscommand(cmdname):
        cmdline = _fixcmdline(cmdline)
        return GenericBrowser(cmdline)
    else:
        return None

def _register_cmdline(cmdline, update_tryorder=0):
    """Register a generic browser for given command line if not already
    registered"""
    if cmdline.lower() in _browsers:
        return True
    browser = _cmdline_browser(cmdline)
    if browser is None:
        return False
    register(cmdline, None, browser, update_tryorder)

def _synthesize(cmdline, update_tryorder=0, register_cmdline=0):
    """Attempt to synthesize a controller base on existing controllers,
    or create a new generic browser for given command line.

    This is useful to create a controller when a user specifies a command
    line to a browser found during module load but not specifically in
    the browser table.

    If a match is found, we can copy a general controller to operate using
    a specific installation of the desired browser.  If no match is found,
    the a temporary generic browser is returned.

    Returns [class, instance] for the browser or [None, None] if the
    browser executable cannot be found at all.

    """

    # Don't accept commands that don't exist
    cmd, args = _splitcommand(cmdline)
    if not _iscommand(cmd):
        return [None, None]

    # Try to adapt an existing registered browser entry, but not
    # for a command line with args because it may differ in action
    if args == '':
        name = os.path.basename(cmd)
        try:
            command = _browsers[name.lower()]
            controller = command[1]
            if controller and name.lower() == controller.basename:
                import copy
                controller = copy.copy(controller)
                controller.name = cmdline
                controller.basename = os.path.basename(cmdline)
                register(cmdline, None, controller, update_tryorder)
                return [None, controller]
        except KeyError:
            pass

    # Unable to synthesize so we create a generic browser
    if register_cmdline:
        _register_cmdline(cmdline, update_tryorder)
    return _browsers[cmdline.lower()]

def _isexecutable(cmd):
    if os.path.isfile(cmd):
        mode = os.stat(cmd)[stat.ST_MODE]
        # Note: All files are considered executable on Windows
        if mode & stat.S_IXUSR or mode & stat.S_IXGRP or mode & stat.S_IXOTH:
            return cmd

def _iscommand(cmd):
    """Return True if cmd is executable or can be found on the executable
    search path."""
    if _isexecutable(cmd):
        return cmd
    path = os.environ.get("PATH")
    if not path:
        return None
    for d in path.split(os.pathsep):
        exe = os.path.join(d, cmd)
        if _isexecutable(exe):
            return exe
    return None

def _splitcommand(cmd):
    """Extract command name and args from command line"""
    if cmd[0] in ('"', "'"):
        name = cmd[1:cmd[1:].find(cmd[0])+1]
        args = cmd[len(name)+2:].strip()
    else:
        parts = cmd.split()
        for i in range(0, len(parts)):
            if parts[i][-1] != '\\':
                name = ' '.join(parts[:i+1]).replace('\\ ', ' ')
                args = ' '.join(parts[i+1:])
                return name, args
        name = ' '.join(parts)
        args = ''
    return name, args

def _fixcmdline(cmdline):
    """Fix a user-given command line to make sure it has %s and & if
    running on Unix"""
    if cmdline.find('%s') == -1:
        cmdline += " '%s'"
    if sys.platform[:3] not in ("win", "os2"):
        if cmdline.rstrip().rfind("&") < len(cmdline.rstrip()) - 1:
            cmdline += " &"
    return cmdline

def _safequote(cmdline, url):
    """Quote given url safely to prevent designing a url that causes the
    browser launch command line to execute something else, and insert
    it into the given command line"""

    if len(cmdline) < 3:
        return ''

    def posix_fix(url):
        """Additional command line processing for posix when passing without
        quotes or with double quotes"""
        if sys.platform[:3] != "win":
            url = url.replace('\\', r'\\').replace('$', r'\$')
            url = url.replace('`', r'\`').replace(';', r'\;')
        return url

    spos = cmdline.rfind('%s')
    quote = cmdline[spos-1]

    # In posix shell invoked by os.system(), single quote can't be within
    # single quotes even if escaped!
    if sys.platform[:3] != "win" and "'" in url and quote == "'":
        cmdline = cmdline[:spos-1] + "%s" + cmdline[spos+3:]
        quote = ''

    if quote == '"':
        url = posix_fix(url)
        url = url.replace('"', r'\"')
    elif quote == "'":
        url = url.replace("'", r"\'")
    elif "'" not in url:
        url = "'" + url + "'"
    else:
        url = posix_fix(url)
        url = url.replace(' ', r"\ ").replace("'", r"\'").replace('"', r'\"')

    return cmdline % url



PROCESS_CREATION_DELAY = 4


# Note: this class is used on Windows also
class GenericBrowser:
    def __init__(self, cmd):
        self.name, self.args = _splitcommand(cmd)
        self.basename = os.path.basename(self.name)

    def open(self, url, new=0, autoraise=1):
        assert "'" not in url
        command = "'%s' %s" % (self.name, self.args)
        rc = os.system(_safequote(command, url))
        return not rc

    def open_new(self, url):
        return self.open(url)


class Netscape:
    "Launcher class for Netscape browsers."
    def __init__(self, name):
        self.name = name
        self.basename = os.path.basename(name)

    def _remote(self, action, autoraise):
        raise_opt = ("-noraise", "-raise")[autoraise]
        cmd = "'%s' %s -remote '%s' >/dev/null 2>&1" % (self.name,
                                                        raise_opt,
                                                        action)
        rc = os.system(cmd)
        if rc:
            import time
            rc = os.system("'%s' &" % self.name)
            if rc:
                return False
            time.sleep(PROCESS_CREATION_DELAY)
            rc = os.system(cmd)
        return not rc

    def open(self, url, new=0, autoraise=1):
        if new:
            cmd = "openURL(%s, new-window)"
        else:
            cmd = "openURL(%s)"
        return self._remote(_safequote(cmd, url), autoraise)

    def open_new(self, url):
        return self.open(url, 1)


class Galeon:
    """Launcher class for Galeon browsers."""
    def __init__(self, name):
        self.name = name
        self.basename = os.path.basename(name)

    def _remote(self, action, autoraise):
        raise_opt = ("--noraise", "")[autoraise]
        cmd = "'%s' %s %s >/dev/null 2>&1 &" % (self.name, raise_opt, action)
        rc = os.system(cmd)
        if rc:
            import time
            rc = os.system("'%s' >/dev/null 2>&1 &" % self.name)
            if rc:
                return False
            time.sleep(PROCESS_CREATION_DELAY)
            rc = os.system(cmd)
        return not rc

    def open(self, url, new=0, autoraise=1):
        if new:
            cmd = "-w '%s'"
        else:
            cmd = "-n '%s'"
        return self._remote(_safequote(cmd, url), autoraise)

    def open_new(self, url):
        return self.open(url, 1)


class Konqueror:
    """Controller for the KDE File Manager (kfm, or Konqueror).

    See http://developer.kde.org/documentation/other/kfmclient.html
    for more information on the Konqueror remote-control interface.

    """
    def __init__(self):
        if _iscommand("konqueror"):
            self.name = self.basename = "konqueror"
        else:
            self.name = self.basename = "kfm"

    def _remote(self, action):
        cmd = "kfmclient %s >/dev/null 2>&1" % action
        rc = os.system(cmd)
        if rc:
            import time
            if self.basename == "konqueror":
                rc = os.system(self.name + " --silent &")
            else:
                rc = os.system(self.name + " -d &")
            if rc:
                return False
            time.sleep(PROCESS_CREATION_DELAY)
            rc = os.system(cmd)
        return not rc

    def open(self, url, new=1, autoraise=1):
        # XXX Currently I know no way to prevent KFM from
        # opening a new win.
        assert "'" not in url
        return self._remote(_safequote("openURL '%s'", url))

    open_new = open


class Grail:
    # There should be a way to maintain a connection to Grail, but the
    # Grail remote control protocol doesn't really allow that at this
    # point.  It probably neverwill!
    def _find_grail_rc(self):
        import glob
        import pwd
        import socket
        import tempfile
        tempdir = os.path.join(tempfile.gettempdir(),
                               ".grail-unix")
        user = pwd.getpwuid(os.getuid())[0]
        filename = os.path.join(tempdir, user + "-*")
        maybes = glob.glob(filename)
        if not maybes:
            return None
        s = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        for fn in maybes:
            # need to PING each one until we find one that's live
            try:
                s.connect(fn)
            except socket.error:
                # no good; attempt to clean it out, but don't fail:
                try:
                    os.unlink(fn)
                except IOError:
                    pass
            else:
                return s

    def _remote(self, action):
        s = self._find_grail_rc()
        if not s:
            return False
        s.send(action)
        s.close()
        return True

    def open(self, url, new=0, autoraise=1):
        if new:
            return self._remote("LOADNEW %s" % url)
        else:
            return self._remote("LOAD %s" % url)

    def open_new(self, url):
        return self.open(url, 1)


class WindowsDefault:
    def open(self, url, new=0, autoraise=1):
        os.startfile(url)
        return True

    def open_new(self, url):
        return self.open(url)


# Adapted from patch submitted to SourceForge by Steven J. Burr
class MacOSX:
    """Launcher class for Aqua browsers on Mac OS X

    Optionally specify a browser name on instantiation.  Note that this
    will not work for Aqua browsers if the user has moved the application
    package after installation.

    If no browser is specified, the default browser, as specified in the
    Internet System Preferences panel, will be used.
    """
    def __init__(self, name):
        self.name = name

    def open(self, url, new=0, autoraise=1):
        assert "'" not in url
        if self.name == "default":
            # User called open, open_new or get without a browser parameter
            script = _safequote('open location "%s"', url)
        else:
            # User called get and chose a browser
            if self.name == "OmniWeb":
                toWindow = ""
            else:
                # Include toWindow parameter of OpenURL command for browsers
                # that support it.  0 == new window; -1 == existing
                toWindow = "toWindow %d" % (new - 1)
            cmd = _safequote('OpenURL "%s"', url)
            script = '''tell application "%s"
                            activate
                            %s %s
                        end tell''' % (self.name, cmd, toWindow)
        # Open pipe to AppleScript through osascript command
        osapipe = os.popen("osascript > /dev/null 2>&1", "w")
        if osapipe is None:
            return False
        # Write script to osascript's stdin
        osapipe.write(script)
        rc = osapipe.close()
        return not rc

    def open_new(self, url):
        return self.open(url, 1)

# XXX problematic:  See OS X support below
#class InternetConfig:
    #def open(self, url, new=0, autoraise=1):
        #ic.launchurl(url)

    #def open_new(self, url):
        #self.open(url)


#
# Platform support for Unix
#

# These are the right tests because all these Unix browsers require either
# a console terminal or an X display to run.

# Prefer X browsers if present
if os.environ.get("DISPLAY"):

    # First, the Netscape series
    if _iscommand("firefox"):
        register("firefox", None, Netscape("firefox"))
    if _iscommand("mozilla"):
        register("mozilla", None, Netscape("mozilla"))
    if _iscommand("netscape"):
        register("netscape", None, Netscape("netscape"))
    # Gnome's Galeon
    if _iscommand("galeon"):
        register("galeon", None, Galeon("galeon"))

    # Skipstone, another Gtk/Mozilla based browser
    if _iscommand("skipstone"):
        register("skipstone", None, GenericBrowser(
            "skipstone '%s' >/dev/null &"))

    # Next, Mosaic -- old but still in use.
    if _iscommand("mosaic"):
        register("mosaic", None, GenericBrowser(
            "mosaic '%s' >/dev/null &"))

    # Konqueror/kfm, the KDE browser.
    if _iscommand("kfm") or _iscommand("konqueror"):
        register("kfm", Konqueror, Konqueror())

    # Grail, the Python browser.
    if _iscommand("grail"):
        register("grail", Grail, None)

# Also try console browsers
if os.environ.get("TERM"):
    # The Links browser <http://artax.karlin.mff.cuni.cz/~mikulas/links/>
    if _iscommand("links"):
        register("links", None, GenericBrowser("links '%s'"))
    # The Lynx browser <http://lynx.browser.org/>
    if _iscommand("lynx"):
        register("lynx", None, GenericBrowser("lynx '%s'"))
    # The w3m browser <http://ei5nazha.yz.yamagata-u.ac.jp/~aito/w3m/eng/>
    if _iscommand("w3m"):
        register("w3m", None, GenericBrowser("w3m '%s'"))

#
# Platform support for Windows
#

if sys.platform[:3] == "win":
    _tryorder = []
    _browsers = {}
    # Prefer netscape if present
    if _iscommand("netscape"):
        register("netscape", None, GenericBrowser("netscape"))
    register("windows-default", WindowsDefault)

#
# Platform support for MacOS
#

if sys.platform == 'darwin':
    # Don't clear _tryorder or _browsers since OS X can use above Unix support
    # (but we prefer using the OS X specific stuff)

    # XXX importing ic will crash entirely in some cases on OS 10.1!
    # XXX how to detect this?
    #try:
        #import ic
    #except ImportError:
        #pass
    #else:
        #register("internet-config", InternetConfig, update_tryorder=-1)

    register("MacOSX", None, MacOSX('default'), -1)

#
# Platform support for OS/2
#

if sys.platform[:3] == "os2" and _iscommand("netscape.exe"):
    _tryorder = []
    _browsers = {}
    register("os2netscape", None,
             GenericBrowser("start netscape.exe %s"), -1)

# OK, now that we know what the default preference orders for each
# platform are, allow user to override them with the BROWSER variable.
if "BROWSER" in os.environ:

    _userchoices = os.environ["BROWSER"].split(os.pathsep)
    _userchoices.reverse()

    # Treat choices in same way as if passed into get() but do register
    # and prepend to _tryorder
    for cmdline in _userchoices:
        _synthesize(cmdline, -1, True)
    cmdline = None # to make del work if _userchoices was empty
    del cmdline
    del _userchoices

# XXX what to do if _tryorder is now empty?
